////////////////////////////////////////////////////////////////
// Intro. to Robotic Algorithms
// Assignment 3
//
// 
////////////////////////////////////////////////////////////////
#include <iostream>

#include "world.h"
#include "mapproblem.h"
#include <math.h> 

//compensate for windows deficiency
#ifndef M_PI
#define M_PI 3.1415926535897932384626433832
#endif


// some globals for use in multiple routines
//
Graphics G;
MapProblem *MP;
objectList workspace, mapGraphics, robotGraphics;
bool workspace_displayed, robot_displayed;
double *readings;     // pointer to storage for sonar readings
int readingsProcessed; // how many readings processed from current scan
int numScans = 0;
double Rx, Ry, Rth;


// "images" for displaying
drawableImage *EmptyMap, *FullMap, *CombinedMap, *EmptyReading, *FullReading;
Bbox gridBbox;

// prototypes
void myKeyboardHandler(unsigned char key);
void makeRandomMap(drawableImage *m, unsigned int xcells, unsigned int ycells);
bool processReading(bool get_next_scan, bool verbose=false);
bool processScan(bool verbose=false);
void integrateCertainty(double rx, double ry, double rth,
			double range, double orient, double rradius);
void integrateDempsterShafer(double rx, double ry, double rth,
			     double range, double orient, double rradius);
void intensifyReadingMaps();
void createCombinedMap();

int main(int argc, char *argv[])
{
  // make sure there are enough arguments
  if (argc < 2) {
    cerr << "Usage: " << argv[0] << " [GLUT options] <problem filename>" <<
      endl;
    exit(1);
  }

  // read in the problem (this must be done before graphics initialization)
  MP = new MapProblem(argv[argc - 1]);

  
  // set up graphics objects
  workspace.push_back(MP->W->boundary);
  copy(MP->W->obstacles.begin(), MP->W->obstacles.end(), 
       back_insert_iterator<objectList>(workspace));
  robotGraphics.push_back(MP->R);

  Bbox worldBbox = MP->W->bbox();
  cout << "word bbox is: " << worldBbox << endl;
  double width = worldBbox.xmax() - worldBbox.xmin();
  double height = worldBbox.ymax() - worldBbox.ymin();
  int xcells = (int) ceil(width/MP->cellSize);
  int ycells = (int) ceil(height/MP->cellSize);
  // calculate the actual grid size based on an integral number of
  // cells that will cover the world
  double xmaximum, ymaximum;
//   cout << "Enter Xmax and Ymax" << endl;
//   cin >> xmaximum >> ymaximum;
  xmaximum = worldBbox.xmin()+xcells*MP->cellSize;
  ymaximum = worldBbox.ymin()+ycells*MP->cellSize;
  gridBbox = Bbox(worldBbox.xmin(), 
		  worldBbox.ymin(),
		  xmaximum,
		  ymaximum);
// 		  worldBbox.xmin()+xcells*MP->cellSize,
// 		  worldBbox.ymin()+ycells*MP->cellSize);
  cout << " here's the measurements" << endl;
  cout << width << " " << height << " " << xcells << " " << ycells << endl;
  cout << gridBbox << endl;

  EmptyMap = new drawableImage(gridBbox, xcells, ycells, drawableImage::BPP1);
  FullMap = new drawableImage(gridBbox, xcells, ycells, drawableImage::BPP1);
  CombinedMap = new drawableImage(gridBbox,xcells,ycells, drawableImage::BPP1);
  EmptyReading = new drawableImage(gridBbox,xcells,ycells,drawableImage::BPP1);
  FullReading = new drawableImage(gridBbox,xcells,ycells,drawableImage::BPP1);
  mapGraphics.push_back(EmptyMap);

  makeRandomMap(EmptyMap, xcells, ycells);
  makeRandomMap(FullMap, xcells, ycells);
  makeRandomMap(FullReading, xcells, ycells);
  makeRandomMap(CombinedMap, xcells, ycells);
  makeRandomMap(EmptyReading, xcells, ycells);

  // initialize variables for storing and handling SONAR readings
  readings = new double[MP->D->numSonars()];
  MP->D->getConfig(&Rx, &Ry, &Rth);
  MP->D->getReadings(readings);	// get first scan
  readingsProcessed = 0;

  // initialize the graphics --- make the window slightly bigger than
  // the world boundary
  const double marginFrac = 0.075;
  Bbox bounds = MP->W->boundary->bbox();
  double Xmargin, Ymargin;
  if( bounds.xmax()-bounds.xmin() >= bounds.ymax()-bounds.ymin() ) {
    Xmargin = marginFrac * (bounds.xmax() - bounds.xmin());
    Ymargin = (bounds.xmax()-bounds.xmin()+ 2*Xmargin 
	       - (bounds.ymax()-bounds.ymin()))/2.0;
  } 
  else {
    Ymargin = marginFrac * (bounds.ymax() - bounds.ymin());
    Xmargin = (bounds.ymax()-bounds.ymin()+2*Ymargin 
	       - (bounds.xmax()-bounds.xmin()))/2.0;
  }

  G.initGL( &argc, argv, "IRA Assignment 3", 640, 640 );
  G.setBackgroundColor("gray50");
  G.setWorldBounds(Bbox(bounds.xmin()-Xmargin, bounds.ymin()-Ymargin,
			bounds.xmax()+Xmargin, bounds.ymax()+Ymargin));
  G.setKeyboardHandler( myKeyboardHandler );
  G.addObjectList(mapGraphics);
  G.addObjectList(workspace);
  workspace_displayed = true;
  G.addObjectList(robotGraphics);
  robot_displayed = true;
  G.runEventLoop();

  return 0;
}

// used for testing; the drawable image must already exist
void makeRandomMap(drawableImage *m, unsigned int xcells, unsigned int ycells)
{
  for (unsigned int x=0; x<xcells; x++)
    for (unsigned int y=0; y<ycells; y++) {
      unsigned char p = rand()%256; // not the "right" way to do it...
      m->setPixel(x, y, &p);
    }
}

void myKeyboardHandler(unsigned char key)
{
  switch (key) {
  case 'E': 
    mapGraphics.clear(); mapGraphics.push_back(EmptyMap); break;
  case 'F':
    mapGraphics.clear(); mapGraphics.push_back(FullMap); break;
  case 'e': 
    mapGraphics.clear(); mapGraphics.push_back(EmptyReading); break;
  case 'f':
    mapGraphics.clear(); mapGraphics.push_back(FullReading); break;
  case 'C':
    mapGraphics.clear(); mapGraphics.push_back(CombinedMap); break;
  case 'N': // no map
    mapGraphics.clear(); break;

  case 'n': // next reading
    processReading(true, true); break;
  case 's': // complete the current scan
    processScan(true);
    break;
  case 't': // complete twenty scans
    {
    for (int k=0; k<20; k++)
      processScan(true);
    }
    break;
  case 'a': // complete all scans
    while (processScan(true));
    break;

  case 'c': 
    createCombinedMap(); break;
  case 'i':
    intensifyReadingMaps(); break;

  case 'w':
    if (workspace_displayed) 
      G.delObjectList(workspace);
    else 
      G.addObjectList(workspace);
    workspace_displayed = !workspace_displayed;
    break;
  case 'r':
    if (robot_displayed) 
      G.delObjectList(robotGraphics);
    else 
      G.addObjectList(robotGraphics);
    robot_displayed = !robot_displayed;
    break;



  case 'h': 
    cout << "Commands: " << endl
	 << "  'h' to print this help listing" << endl
	 << "  'q' to quit the program" << endl
	 << endl
	 << "  'n' to process the next reading " << endl
	 << "  's' to process the next scan" << endl
	 << "  't' to process the next ten scans" << endl
	 << "  'a' to process all remaining readings" << endl
	 << endl 
	 << "  'e' to display the ``empty'' sonar reading map" << endl
	 << "  'f' to display the ``full'' sonar reading map" << endl
	 << "  'i' to intensify the nonzero sonar reading map elements" << endl
	 << "  'E' to display the ``Empty'' map" << endl
	 << "  'F' to display the ``Full'' map" << endl
	 << "  'C' to display a ``combined'' map" << endl
	 << endl
	 << "  'c' to create the combined map" << endl
	 << "  'w' to toggle display of the workspace" << endl
	 << "  'd' to dump the current map to a file" << endl

	 << endl;
    G.printControlKeyBindings();
    break;
  case 'q': 
    exit(0);
  }
}

// process all the (remaining) readings from a single sonar scan
//
// return true if the end of file has not been encountered
// 
bool processScan(bool verbose)
{
  // process the next reading; if we've finished with the current
  // scan, then get the next one
  if (processReading(true, verbose))
    while (processReading(false, verbose));
  return !MP->D->endOfFile();
}

// void processReading(bool verbose=false)
// {
//   double x, y, th;
//   MP->D->getConfig(&x, &y, &th);
// //   processReading(x, y, th, readings[readingsProcessed], verbose);
// }


// process one reading from the current scan
// 
// if there are no more readings from the current scan (or if we've
// processed them all already) then return false
// 
bool processReading(bool get_next_scan, bool verbose)
{
  if (readingsProcessed >= MP->D->numSonars()) {
    if (MP->D->endOfFile()) {
      if (verbose)
	cerr << endl << "No more sonar data in file!" << endl << endl;
      return false;
    }
    else if (!get_next_scan) {
      if (verbose)
	cerr << endl <<"No more readings in current scan!" << endl << endl;
      return false;
    }
    else {
      if (MP->D->next())
	return false;
      MP->D->getReadings(readings);
      readingsProcessed = 0;
      MP->D->getConfig(&Rx, &Ry, &Rth);
    }
  }
  if (verbose && readingsProcessed == 0)
    cout << endl << "Position is x=" << Rx << " y=" << Ry << endl
	 << "Processing scan #" << numScans++ << endl;
  if (verbose) 
    cout << endl << "Processing reading #" << readingsProcessed << endl;
  
  // integrate the reading into the map
  if (MP->technique == "certainty")
    integrateCertainty(Rx, Ry, Rth, readings[readingsProcessed], 
		       MP->D->sonarOrientation(readingsProcessed),
		       MP->D->ringRadius());
  else if (MP->technique == "Dempster-Shafer")
    integrateDempsterShafer(Rx, Ry, Rth, readings[readingsProcessed], 
		       MP->D->sonarOrientation(readingsProcessed),
		       MP->D->ringRadius());
  else {
    cerr << "The requested \"technique\" must be either \"certainty\""
	 << " or \"Dempster-Shafer\"" << endl;
    return false;
  }

  readingsProcessed++;

  return true;
}


// This function "destructively" modifies the EmptyReading and
// FullReading maps by setting any nonzero value to 255.  This is
// useful to see which pixels you have selected to be in the sonar
// cone.
// 
void intensifyReadingMaps()
{
  const unsigned char v = 255;
  const unsigned char z = 0;

  for (unsigned int x=0; x < EmptyReading->getWidth(); x++) 
    for (unsigned int y=0; y < EmptyReading->getHeight(); y++) {
      if (*(EmptyReading->getPixel(x,y)) > 128)
//       if (*(EmptyReading->getPixel(x,y)) != 0)
	EmptyReading->setPixel(x,y,&v);
      else
	EmptyReading->setPixel(x,y,&z);

//       if (*(FullReading->getPixel(x,y)) != 0)
      if (*(FullReading->getPixel(x,y)) > 128)
	FullReading->setPixel(x,y,&v);
      else
	FullReading->setPixel(x,y,&z);
    }
}


////////////////////////////////////////////////////////////////
// YOUR CODE HERE: integrateCertainty() and integrateDempsterShafer()
//
// These procedures are passed the configuration of the robot (rx, ry,
// rth), the range returned by the sonar sensor, the orientation of
// the sensor axis, and the radius of the sonar ring "rradius".
//
// They should set the values of the EmptyReading and FullReading maps
// to reflect the model of the sonar sensor and then merge these
// readings into the EmptyMap and FullMap maps.
//
// They should not update the CombinedMap.
// 
void integrateCertainty(double rx, double ry, double rth,
			double range, double orient, double rradius)
{
}


void integrateDempsterShafer(double rx, double ry, double rth,
			     double range, double orient, double rradius)
{
}

////////////////////////////////////////////////////////////////
// YOUR CODE HERE: createCombinedMap()
//
void createCombinedMap()
{
}
