#ifndef MAPPROBLEM_H
#define MAPPROBLEM_H

#include <filereader.h>
#include <string.h>
#include "world.h"

class SonarData;

class MapProblem {
private:

public:
  World *W;
  Robot *R;
  SonarData *D;
  string problemName;
  string technique; // should be either "uncertainty" or "dempster-shafer"
  double cellSize;

  MapProblem(const char* fname);

};


class SonarData {
private:
  ifstream *F;
  int _numSonars;
  double *sonarOrientations;
  double sonarRingRadius;
  bool doneWithFile;

  double *currentReadings;
  double x, y, th;

public:
  SonarData(const char* fname, bool verbose=false);

  int numSonars() const { return _numSonars; }
  double ringRadius() const { return sonarRingRadius; } // in meters

  // returns orientation of given sonar (zero indexed) in radians
  double sonarOrientation(int j) const { return sonarOrientations[j]; }

  // returns the current position (in meters) and the orientation (in radians)
  void getConfig(double *x, double *y, double *th);
		   
  // get the current sonar readings
  // "readings" must be an existing array with room for "numSonars" elements
  void getReadings(double *readings);

  // set the current sonar readings to the next scan in the file
  // returns "true" if successful (i.e. end of file not reached)
  bool next(bool verbose=false);
  bool endOfFile() const { return doneWithFile; }

  ~SonarData() 
  { 
    delete[] sonarOrientations; 
    delete[] currentReadings; 
    F->close();
  }
};



#endif
